@@ -25,11 +25,18 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <errno.h>
+#include <stdio.h>
+#include <mntent.h>
+
 
 #include "xdp-utils.h"
 
+#define DBUS_NAME_DBUS "org.freedesktop.DBus"
+#define DBUS_INTERFACE_DBUS DBUS_NAME_DBUS
+#define DBUS_PATH_DBUS "/org/freedesktop/DBus"
+
 G_LOCK_DEFINE (app_infos);
-static GHashTable *app_infos;
+static GHashTable *app_info_by_unique_name;
 
 /* Based on g_mkstemp from glib */
 
@@ -96,24 +103,44 @@ xdp_mkstempat (int    dir_fd,
   return -1;
 }
 
+typedef enum
+{
+  XDP_APP_INFO_KIND_HOST = 0,
+  XDP_APP_INFO_KIND_FLATPAK = 1,
+  XDP_APP_INFO_KIND_SNAP    = 2,
+} XdpAppInfoKind;
+
 struct _XdpAppInfo {
   volatile gint ref_count;
   char *id;
-  GKeyFile *flatpak_metadata;
+  XdpAppInfoKind kind;
+
+  union
+    {
+      struct
+        {
+          GKeyFile *keyfile;
+        } flatpak;
+      struct
+        {
+          int dummy;
+        } snap;
+    } u;
 };
 
 static XdpAppInfo *
-xdp_app_info_new (void)
+xdp_app_info_new (XdpAppInfoKind kind)
 {
   XdpAppInfo *app_info = g_new0 (XdpAppInfo, 1);
   app_info->ref_count = 1;
+  app_info->kind = kind;
   return app_info;
 }
 
 static XdpAppInfo *
 xdp_app_info_new_host (void)
 {
-  XdpAppInfo *app_info = xdp_app_info_new ();
+  XdpAppInfo *app_info = xdp_app_info_new (XDP_APP_INFO_KIND_HOST);
   app_info->id = g_strdup ("");
   return app_info;
 }
@@ -122,7 +149,21 @@ static void
 xdp_app_info_free (XdpAppInfo *app_info)
 {
   g_free (app_info->id);
-  g_clear_pointer (&app_info->flatpak_metadata, g_key_file_free);
+
+  switch (app_info->kind)
+    {
+    case XDP_APP_INFO_KIND_FLATPAK:
+      g_clear_pointer (&app_info->u.flatpak.keyfile, g_key_file_free);
+      break;
+
+    case XDP_APP_INFO_KIND_SNAP:
+      break;
+
+    case XDP_APP_INFO_KIND_HOST:
+    default:
+      break;
+    }
+
   g_free (app_info);
 }
 
@@ -157,19 +198,27 @@ xdp_app_info_is_host (XdpAppInfo *app_info)
 {
   g_return_val_if_fail (app_info != NULL, FALSE);
 
-  return strcmp (app_info->id, "") == 0;
+  return app_info->kind == XDP_APP_INFO_KIND_HOST;
+}
+
+gboolean
+xdp_app_info_supports_opath (XdpAppInfo  *app_info)
+{
+  return
+    app_info->kind == XDP_APP_INFO_KIND_FLATPAK ||
+    app_info->kind == XDP_APP_INFO_KIND_HOST;
 }
 
 char *
 xdp_app_info_remap_path (XdpAppInfo *app_info,
                          const char *path)
 {
-  if (app_info->flatpak_metadata)
+  if (app_info->kind == XDP_APP_INFO_KIND_FLATPAK)
     {
-      g_autofree char *app_path = g_key_file_get_string (app_info->flatpak_metadata,
+      g_autofree char *app_path = g_key_file_get_string (app_info->u.flatpak.keyfile,
                                                          FLATPAK_METADATA_GROUP_INSTANCE,
                                                          FLATPAK_METADATA_KEY_APP_PATH, NULL);
-      g_autofree char *runtime_path = g_key_file_get_string (app_info->flatpak_metadata,
+      g_autofree char *runtime_path = g_key_file_get_string (app_info->u.flatpak.keyfile,
                                                              FLATPAK_METADATA_GROUP_INSTANCE,
                                                              FLATPAK_METADATA_KEY_RUNTIME_PATH,
                                                              NULL);
@@ -198,14 +247,15 @@ xdp_app_info_remap_path (XdpAppInfo *app_info,
 }
 
 static void
-ensure_app_infos (void)
+ensure_app_info_by_unique_name (void)
 {
-  if (app_infos == NULL)
-    app_infos = g_hash_table_new_full (g_str_hash, g_str_equal,
-                                       g_free, (GDestroyNotify)xdp_app_info_unref);
+  if (app_info_by_unique_name == NULL)
+    app_info_by_unique_name = g_hash_table_new_full (g_str_hash, g_str_equal,
+                                                     g_free,
+                                                     (GDestroyNotify)xdp_app_info_unref);
 }
 
-/* Returns NULL on failure, keyfile with name "" if not sandboxed, and full app-info otherwise */
+/* Returns NULL with error set on failure, NULL with no error set if not a flatpak, and app-info otherwise */
 static XdpAppInfo *
 parse_app_info_from_flatpak_info (int pid, GError **error)
 {
@@ -242,8 +292,8 @@ parse_app_info_from_flatpak_info (int pid, GError **error)
     {
       if (errno == ENOENT)
         {
-          /* No file => on the host */
-          return xdp_app_info_new_host ();
+          /* No file => on the host, return NULL with no error */
+          return NULL;
         }
 
       /* Some weird error => failure */
@@ -289,22 +339,161 @@ parse_app_info_from_flatpak_info (int pid, GError **error)
   if (id == NULL)
     return NULL;
 
-  app_info = xdp_app_info_new ();
+  app_info = xdp_app_info_new (XDP_APP_INFO_KIND_FLATPAK);
   app_info->id = g_steal_pointer (&id);
-  app_info->flatpak_metadata = g_steal_pointer (&metadata);
+  app_info->u.flatpak.keyfile = g_steal_pointer (&metadata);
 
   return g_steal_pointer (&app_info);
 }
 
+static gboolean
+aa_is_enabled (void)
+{
+  static int apparmor_enabled = -1;
+  struct stat statbuf;
+  struct mntent *mntpt;
+  FILE *mntfile;
+
+  if (apparmor_enabled >= 0)
+    return apparmor_enabled;
+
+  apparmor_enabled = FALSE;
+
+  mntfile = setmntent ("/proc/mounts", "r");
+  if (!mntfile)
+    return FALSE;
+
+  while ((mntpt = getmntent (mntfile)))
+    {
+      g_autofree char *proposed = NULL;
+
+      if (strcmp (mntpt->mnt_type, "securityfs") != 0)
+        continue;
+
+      proposed = g_strdup_printf ("%s/apparmor", mntpt->mnt_dir);
+      if (stat (proposed, &statbuf) == 0)
+        {
+          apparmor_enabled = TRUE;
+          break;
+        }
+    }
+
+  endmntent (mntfile);
+
+  return apparmor_enabled;
+}
+
+#define UNCONFINED		"unconfined"
+#define UNCONFINED_SIZE		strlen(UNCONFINED)
+
+static gboolean
+parse_unconfined (char *con, int size)
+{
+  return size == UNCONFINED_SIZE && strncmp (con, UNCONFINED, UNCONFINED_SIZE) == 0;
+}
+
+static char *
+aa_splitcon (char *con, char **mode)
+{
+  char *label = NULL;
+  char *mode_str = NULL;
+  char *newline = NULL;
+  int size = strlen (con);
+
+  if (size == 0)
+    return NULL;
+
+  /* Strip newline */
+  if (con[size - 1] == '\n')
+    {
+      newline = &con[size - 1];
+      size--;
+    }
+
+  if (parse_unconfined (con, size))
+    {
+      label = con;
+    }
+  else if (size > 3 && con[size - 1] == ')')
+    {
+      int pos = size - 2;
+
+      while (pos > 0 && !(con[pos] == ' ' && con[pos + 1] == '('))
+        pos--;
+
+      if (pos > 0)
+        {
+          con[pos] = 0; /* overwrite ' ' */
+          con[size - 1] = 0; /* overwrite trailing ) */
+          mode_str = &con[pos + 2]; /* skip '(' */
+          label = con;
+        }
+    }
+
+  if (label && newline)
+    *newline = 0; /* overwrite '\n', if requested, on success */
+  if (mode)
+    *mode = mode_str;
+
+  return label;
+}
+
+static XdpAppInfo *
+parse_app_info_from_security_label (const char *security_label)
+{
+  char *label, *dot;
+  g_autofree char *snap_name = NULL;
+  g_autoptr(XdpAppInfo) app_info = NULL;
+
+  /* Snap confinement requires AppArmor */
+  if (aa_is_enabled ())
+    {
+      /* Parse the security label as an AppArmor context.  We take a copy
+       * of the string because aa_splitcon modifies its argument. */
+      g_autofree char *security_label_copy = g_strdup (security_label);
+
+      label = aa_splitcon (security_label_copy, NULL);
+      if (label && g_str_has_prefix (label, "snap."))
+        {
+          /* If the label belongs to a snap, it will be of the form
+           * snap.$PACKAGE.$APPLICATION.  We want to extract the package
+           * name */
+
+          label += 5;
+          dot = strchr (label, '.');
+          if (!dot)
+            return NULL;
+          snap_name = g_strndup (label, dot - label);
+
+          app_info = xdp_app_info_new (XDP_APP_INFO_KIND_SNAP);
+          app_info->id = g_strconcat ("snap.", snap_name, NULL);
+
+          return g_steal_pointer (&app_info);
+        }
+    }
+
+  return NULL;
+}
+
+
 XdpAppInfo *
 xdp_get_app_info_from_pid (pid_t pid,
                            GError **error)
 {
   g_autoptr(XdpAppInfo) app_info = NULL;
+  g_autoptr(GError) local_error = NULL;
+
+  /* TODO: Handle snap support via apparmor here */
+
+  app_info = parse_app_info_from_flatpak_info (pid, &local_error);
+  if (app_info == NULL && local_error)
+    {
+      g_propagate_error (error, g_steal_pointer (&local_error));
+      return NULL;
+    }
 
-  app_info = parse_app_info_from_flatpak_info (pid, error);
   if (app_info == NULL)
-    return NULL;
+    app_info = xdp_app_info_new_host ();
 
   return app_info;
 }
@@ -315,9 +504,9 @@ lookup_cached_app_info_by_sender (const char *sender)
   XdpAppInfo *app_info = NULL;
 
   G_LOCK (app_infos);
-  if (app_infos)
+  if (app_info_by_unique_name)
     {
-      app_info = g_hash_table_lookup (app_infos, sender);
+      app_info = g_hash_table_lookup (app_info_by_unique_name, sender);
       if (app_info)
         xdp_app_info_ref (app_info);
     }
@@ -339,15 +528,16 @@ xdp_connection_lookup_app_info_sync (GDBusConnection       *connection,
   g_autoptr(GVariantIter) iter = NULL;
   const char *key;
   GVariant *value;
+  g_autofree char *security_label = NULL;
   guint32 pid = 0;
 
   app_info = lookup_cached_app_info_by_sender (sender);
   if (app_info)
     return g_steal_pointer (&app_info);
 
-  msg = g_dbus_message_new_method_call ("org.freedesktop.DBus",
-                                        "/org/freedesktop/DBus",
-                                        "org.freedesktop.DBus",
+  msg = g_dbus_message_new_method_call (DBUS_NAME_DBUS,
+                                        DBUS_PATH_DBUS,
+                                        DBUS_INTERFACE_DBUS,
                                         "GetConnectionCredentials");
   g_dbus_message_set_body (msg, g_variant_new ("(s)", sender));
 
@@ -373,15 +563,36 @@ xdp_connection_lookup_app_info_sync (GDBusConnection       *connection,
     {
       if (strcmp (key, "ProcessID") == 0)
         pid = g_variant_get_uint32 (value);
+      else if (strcmp (key, "LinuxSecurityLabel") == 0)
+        {
+          g_clear_pointer (&security_label, g_free);
+          security_label = g_variant_dup_bytestring (value, NULL);
+        }
+    }
+
+  if (app_info == NULL && security_label != NULL)
+    {
+      app_info = parse_app_info_from_security_label (security_label);
     }
 
-  app_info = parse_app_info_from_flatpak_info (pid, error);
   if (app_info == NULL)
-    return NULL;
+    {
+      g_autoptr(GError) local_error = NULL;
+      app_info = parse_app_info_from_flatpak_info (pid, &local_error);
+      if (app_info == NULL && local_error)
+        {
+          g_propagate_error (error, g_steal_pointer (&local_error));
+          return NULL;
+        }
+    }
+
+  if (app_info == NULL)
+    app_info = xdp_app_info_new_host ();
 
   G_LOCK (app_infos);
-  ensure_app_infos ();
-  g_hash_table_insert (app_infos, g_strdup (sender), xdp_app_info_ref (app_info));
+  ensure_app_info_by_unique_name ();
+  g_hash_table_insert (app_info_by_unique_name, g_strdup (sender),
+                       xdp_app_info_ref (app_info));
   G_UNLOCK (app_infos);
 
   return g_steal_pointer (&app_info);
@@ -410,15 +621,15 @@ name_owner_changed (GDBusConnection *connection,
   const char *name, *from, *to;
   XdpPeerDiedCallback peer_died_cb = user_data;
 
-  g_variant_get (parameters, "(sss)", &name, &from, &to);
+  g_variant_get (parameters, "(&s&s&s)", &name, &from, &to);
 
   if (name[0] == ':' &&
       strcmp (name, from) == 0 &&
       strcmp (to, "") == 0)
     {
       G_LOCK (app_infos);
-      if (app_infos)
-        g_hash_table_remove (app_infos, name);
+      if (app_info_by_unique_name)
+        g_hash_table_remove (app_info_by_unique_name, name);
       G_UNLOCK (app_infos);
 
       if (peer_died_cb)
@@ -431,10 +642,10 @@ xdp_connection_track_name_owners (GDBusConnection *connection,
                                   XdpPeerDiedCallback peer_died_cb)
 {
   g_dbus_connection_signal_subscribe (connection,
-                                      "org.freedesktop.DBus",
-                                      "org.freedesktop.DBus",
+                                      DBUS_NAME_DBUS,
+                                      DBUS_INTERFACE_DBUS,
                                       "NameOwnerChanged",
-                                      "/org/freedesktop/DBus",
+                                      DBUS_PATH_DBUS,
                                       NULL,
                                       G_DBUS_SIGNAL_FLAGS_NONE,
                                       name_owner_changed,
@@ -482,66 +693,160 @@ xdg_desktop_portal_error_quark (void)
   return (GQuark) quark_volatile;
 }
 
-char *
-xdp_app_info_get_path_for_fd (XdpAppInfo *app_info,
-                              int fd)
+static char *
+verify_proc_self_fd (XdpAppInfo *app_info,
+                     const char *proc_path)
 {
-  g_autofree char *proc_path = NULL;
-  int fd_flags;
   char path_buffer[PATH_MAX + 1];
-  g_autofree char *path = NULL;
-  ssize_t symlink_size = 0;
-  struct stat st_buf;
-  struct stat real_st_buf;
+  ssize_t symlink_size;
 
-  proc_path = g_strdup_printf ("/proc/self/fd/%d", fd);
-
-  if (fd == -1 ||
-      (fd_flags = fcntl (fd, F_GETFL)) == 0 ||
-      ((fd_flags & O_PATH) != O_PATH) ||
-      ((fd_flags & O_NOFOLLOW) == O_NOFOLLOW) ||
-      fstat (fd, &st_buf) < 0 ||
-      ((st_buf.st_mode & S_IFMT) != S_IFREG && (st_buf.st_mode & S_IFMT) != S_IFDIR) ||
-      (symlink_size = readlink (proc_path, path_buffer, PATH_MAX)) < 0)
-    {
-      return NULL;
-    }
+  symlink_size = readlink (proc_path, path_buffer, PATH_MAX);
+  if (symlink_size < 0)
+    return NULL;
 
   path_buffer[symlink_size] = 0;
 
-  path = xdp_app_info_remap_path (app_info, path_buffer);
+  /* All normal paths start with /, but some weird things
+     don't, such as socket:[27345] or anon_inode:[eventfd].
+     We don't support any of these */
+  if (path_buffer[0] != '/')
+    return NULL;
+
+  /* File descriptors to actually deleted files have " (deleted)"
+     appended to them. This also happens to some fake fd types
+     like shmem which are "/<name> (deleted)". All such
+     files are considered invalid. Unfortunatelly this also
+     matches files with filenames that actually end in " (deleted)",
+     but there is not much to do about this. */
+  if (g_str_has_suffix (path_buffer, " (deleted)"))
+    return NULL;
+
+  /* remap from sandbox to host if needed */
+  return xdp_app_info_remap_path (app_info, path_buffer);
+}
+
+char *
+xdp_app_info_get_path_for_fd (XdpAppInfo *app_info,
+                              int fd,
+                              int require_st_mode,
+                              struct stat *st_buf,
+                              gboolean *writable_out)
+{
+  g_autofree char *proc_path = NULL;
+  int fd_flags;
+  struct stat st_buf_store;
+  struct stat real_st_buf;
+  gboolean writable = FALSE;
+  g_autofree char *path = NULL;
+
+  if (st_buf == NULL)
+    st_buf = &st_buf_store;
+
+  if (fd == -1)
+    return NULL;
+
+  /* Must be able to get fd flags */
+  fd_flags = fcntl (fd, F_GETFL);
+  if (fd_flags == -1)
+    return NULL;
+
+  /* Must be able to fstat */
+  if (fstat (fd, st_buf) < 0)
+    return NULL;
+
+  /* Verify mode */
+  if (require_st_mode != 0 &&
+      (st_buf->st_mode & S_IFMT) != require_st_mode)
+    return NULL;
+
+  proc_path = g_strdup_printf ("/proc/self/fd/%d", fd);
+
+  /* Must be able to read valid path from /proc/self/fd */
+  /* This is an absolute and (at least at open time) symlink-expanded path */
+  path = verify_proc_self_fd (app_info, proc_path);
+  if (path == NULL)
+    return NULL;
+
+  if ((fd_flags & O_PATH) == O_PATH)
+    {
+      int read_access_mode;
+
+      /* Earlier versions of the portal supported only O_PATH fds, as
+       * these are safer to handle on the portal side. But we now
+       * prefer regular FDs because these ensure that the sandbox
+       * actually has full access to the file in its security context.
+       *
+       * However, we still support O_PATH fds when possible because
+       * existing code uses it.
+       *
+       * See issues #167 for details.
+       */
+
+      /* Must not be O_NOFOLLOW (because we want the target file) */
+      if ((fd_flags & O_NOFOLLOW) == O_NOFOLLOW)
+        return NULL;
+
+      if (!xdp_app_info_supports_opath (app_info))
+        return NULL;
+
+      read_access_mode = R_OK;
+      if (S_ISDIR (st_buf->st_mode))
+        read_access_mode |= X_OK;
+
+      /* Must be able to access the path via the sandbox supplied O_PATH fd,
+         which applies the sandbox side mount options (like readonly). */
+      if (access (proc_path, read_access_mode) != 0)
+        return NULL;
+
+      if (xdp_app_info_is_host (app_info) || access (proc_path, W_OK) == 0)
+        writable = TRUE;
+    }
+  else /* Regular file with no O_PATH */
+    {
+      int accmode = fd_flags & O_ACCMODE;
+
+      /* Note that this only gives valid results for writable for regular files,
+         as there is no way to get a writable fd for a directory. */
+
+      /* Don't allow WRONLY (or weird) open modes */
+      if (accmode != O_RDONLY &&
+          accmode != O_RDWR)
+        return NULL;
+
+      if (xdp_app_info_is_host (app_info) || accmode == O_RDWR)
+        writable = TRUE;
+    }
 
   /* Verify that this is the same file as the app opened */
   if (stat (path, &real_st_buf) < 0 ||
-      st_buf.st_dev != real_st_buf.st_dev ||
-      st_buf.st_ino != real_st_buf.st_ino)
+      st_buf->st_dev != real_st_buf.st_dev ||
+      st_buf->st_ino != real_st_buf.st_ino)
     {
       /* Different files on the inside and the outside, reject the request */
       return NULL;
     }
 
-  return g_steal_pointer (&path);
-}
+  if (writable_out)
+    *writable_out = writable;
 
-static gboolean
-is_valid_initial_name_character (gint c, gboolean allow_dash)
-{
-  return
-    (c >= 'A' && c <= 'Z') ||
-    (c >= 'a' && c <= 'z') ||
-    (c == '_') || (allow_dash && c == '-');
+  return g_steal_pointer (&path);
 }
 
 static gboolean
 is_valid_name_character (gint c, gboolean allow_dash)
 {
   return
-    is_valid_initial_name_character (c, allow_dash) ||
-    (c >= '0' && c <= '9');
+    (c >= 'A' && c <= 'Z') ||
+    (c >= 'a' && c <= 'z') ||
+    (c >= '0' && c <= '9') ||
+    (c == '_') || (allow_dash && c == '-');
 }
 
+/* This is the same as flatpak apps, except we also allow
+   names to start with digits, and two-element names so that ids of the form
+   snap.$snapname is allowed for all snap names. */
 gboolean
-xdp_is_valid_flatpak_name (const char *string)
+xdp_is_valid_app_id (const char *string)
 {
   guint len;
   const gchar *s;
@@ -567,10 +872,7 @@ xdp_is_valid_flatpak_name (const char *string)
   s = string;
   if (G_UNLIKELY (*s == '.'))
     return FALSE; /* Name can't start with a period */
-  else if (G_UNLIKELY (!is_valid_initial_name_character (*s, last_element)))
-    return FALSE;
 
-  s += 1;
   dot_count = 0;
   while (s != end)
     {
@@ -581,16 +883,15 @@ xdp_is_valid_flatpak_name (const char *string)
           s += 1;
           if (G_UNLIKELY (s == end))
             return FALSE;
-          if (!is_valid_initial_name_character (*s, last_element))
-            return FALSE;
           dot_count++;
         }
-      else if (G_UNLIKELY (!is_valid_name_character (*s, last_element)))
+
+      if (G_UNLIKELY (!is_valid_name_character (*s, last_element)))
         return FALSE;
       s += 1;
     }
 
-  if (G_UNLIKELY (dot_count < 2))
+  if (G_UNLIKELY (dot_count < 1))
     return FALSE;
 
   return TRUE;
